"use strict";
/**
 * @license
 * Copyright 2018 Google Inc. All Rights Reserved.
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * =============================================================================
 */
var __extends = (this && this.__extends) || (function () {
    var extendStatics = function (d, b) {
        extendStatics = Object.setPrototypeOf ||
            ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
            function (d, b) { for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p]; };
        return extendStatics(d, b);
    };
    return function (d, b) {
        extendStatics(d, b);
        function __() { this.constructor = d; }
        d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
    };
})();
Object.defineProperty(exports, "__esModule", { value: true });
var engine_1 = require("../engine");
var globals_1 = require("../globals");
var ops_1 = require("../ops/ops");
var serialization_1 = require("../serialization");
var optimizer_1 = require("./optimizer");
/** @doclink Optimizer */
var AdadeltaOptimizer = /** @class */ (function (_super) {
    __extends(AdadeltaOptimizer, _super);
    function AdadeltaOptimizer(learningRate, rho, epsilon) {
        if (epsilon === void 0) { epsilon = null; }
        var _this = _super.call(this) || this;
        _this.learningRate = learningRate;
        _this.rho = rho;
        _this.epsilon = epsilon;
        _this.accumulatedGrads = [];
        _this.accumulatedUpdates = [];
        if (epsilon == null) {
            _this.epsilon = engine_1.ENGINE.backend.epsilon();
        }
        return _this;
    }
    AdadeltaOptimizer.prototype.applyGradients = function (variableGradients) {
        var _this = this;
        var variableNames = Array.isArray(variableGradients) ?
            variableGradients.map(function (item) { return item.name; }) :
            Object.keys(variableGradients);
        var _loop_1 = function (i) {
            var name_1 = variableNames[i];
            var value = engine_1.ENGINE.registeredVariables[name_1];
            var trainable = false;
            if (this_1.accumulatedGrads[i] == null) {
                this_1.accumulatedGrads[i] = {
                    name: name_1 + "/accum_grad",
                    variable: globals_1.tidy(function () { return ops_1.zerosLike(value).variable(trainable); })
                };
            }
            if (this_1.accumulatedUpdates[i] == null) {
                this_1.accumulatedUpdates[i] = {
                    name: name_1 + "/accum_var",
                    variable: globals_1.tidy(function () { return ops_1.zerosLike(value).variable(trainable); })
                };
            }
            var gradient = Array.isArray(variableGradients) ?
                variableGradients[i].tensor : variableGradients[name_1];
            var accumulatedGrad = this_1.accumulatedGrads[i];
            var accumulatedUpdate = this_1.accumulatedUpdates[i];
            globals_1.tidy(function () {
                var newAccumulatedGrad = accumulatedGrad.variable.mul(_this.rho).add(gradient.square().mul(1 - _this.rho));
                var updates = accumulatedUpdate.variable.add(_this.epsilon)
                    .sqrt()
                    .div(accumulatedGrad.variable.add(_this.epsilon).sqrt())
                    .mul(gradient);
                var newAccumulatedUpdate = accumulatedUpdate.variable.mul(_this.rho).add(updates.square().mul(1 - _this.rho));
                _this.accumulatedGrads[i].variable.assign(newAccumulatedGrad);
                _this.accumulatedUpdates[i].variable.assign(newAccumulatedUpdate);
                var newValue = updates.mul(-_this.learningRate).add(value);
                value.assign(newValue);
            });
        };
        var this_1 = this;
        for (var i = 0; i < variableNames.length; ++i) {
            _loop_1(i);
        }
    };
    AdadeltaOptimizer.prototype.dispose = function () {
        if (this.accumulatedUpdates != null) {
            globals_1.dispose([
                this.accumulatedGrads.map(function (v) { return v.variable; }),
                this.accumulatedUpdates.map(function (v) { return v.variable; })
            ]);
        }
    };
    AdadeltaOptimizer.prototype.getWeights = function () {
        // Order matters for Python compatibility.
        var namedVariables = this.accumulatedGrads.concat(this.accumulatedUpdates);
        return namedVariables.map(function (v) { return ({ name: v.name, tensor: v.variable }); });
    };
    AdadeltaOptimizer.prototype.setWeights = function (weightValues) {
        var variableCount = weightValues.length / 2;
        var trainable = false;
        this.accumulatedGrads = weightValues.slice(0, variableCount).map(function (v) { return ({ name: v.name, variable: v.tensor.variable(trainable) }); });
        this.accumulatedUpdates =
            weightValues.slice(variableCount, variableCount * 2).map(function (v) { return ({ name: v.name, variable: v.tensor.variable(trainable) }); });
    };
    AdadeltaOptimizer.prototype.getConfig = function () {
        return {
            'learningRate': this.learningRate,
            'rho': this.rho,
            'epsilon': this.epsilon
        };
    };
    /** @nocollapse */
    AdadeltaOptimizer.fromConfig = function (cls, config) {
        return new cls(config['learningRate'], config['rho'], config['epsilon']);
    };
    /** @nocollapse */
    AdadeltaOptimizer.className = 'AdadeltaOptimizer';
    return AdadeltaOptimizer;
}(optimizer_1.Optimizer));
exports.AdadeltaOptimizer = AdadeltaOptimizer;
serialization_1.registerClass(AdadeltaOptimizer);
//# sourceMappingURL=adadelta_optimizer.js.map